#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(WIN32)
#include <unistd.h>
#include <termios.h>
#include <arpa/inet.h>
#else
#include <winsock2.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


#include "shoehorn.h"
#include "operation.h"
#include "serial.h"
#include "log.h"

#define ERRLINE (__LINE__)

#define MAX_FILE_SIZE (0x1f0000)

static char msg[256];

static void
dump_buffer(const char *label, u8 *buf, u32 count)
{
#if defined(DEBUG_THIS)
#if !defined(WIN32)
	u32 i;
	printf("Dump Buffer: %s\n", label);
	for (i=0; i<count; i++) {
		printf("%02x ", buf[i]);
		if (i%16 == 15)
			printf("\n");
	}
	if (i%16 != 0)
		printf("\n");
#else
	int ptr = 0;
	u32 i;

	ptr += sprintf(msg, "Dump Buffer: %s\n", label);
	for (i=0; i<count; i++) {
		ptr += sprintf(&msg[ptr], "%02x ", buf[i]);
		if (i%16 == 15)
			ptr += sprintf(&msg[ptr], "\n");
	}
	message(msg);
#endif
#endif
}

static int
send_cmd(u16 header, u32 addr, u8 format, u32 count, u32 data, u8 end)
{
	u32 cmd_interval = 1;
  
	header = htons(header);
	addr = htonl(addr);
	count = htonl(count);
	data = htonl(data);

	write_serial((u8 *)&header, 2, cmd_interval);
	write_serial((u8 *)&addr, 4, cmd_interval);
	write_serial((u8 *)&format, 1, cmd_interval);
	write_serial((u8 *)&count, 4, cmd_interval);
	write_serial((u8 *)&data, 4, cmd_interval);
	write_serial((u8 *)&end, 1, cmd_interval);

	return 0;
}

static int
wait_response(u8 *buf, u32 count, u32 timeout_ms)
{
	int ret;

	ret = read_serial(buf, count, timeout_ms);
	if ((u32)ret != count)
		return -ERRLINE;
	
	return ret;
}

static int
target_load_image(u8 *data, u32 size)
{
	u8 buf[256];
	int ret;
	u8 response[4] = {0x88, 0x88, 0x88, 0x88};

	send_cmd(0x0404, 0x1fffc000, 0x00, size, 0x00000000, 0xaa);
	ret = wait_response(buf, 4, 10);
	if (ret < 0) {
		return -ERRLINE;
	}
	dump_buffer("Load CMD", buf, 4);

	write_serial(data, size, 1);

	send_cmd(0x0505, 0x00000000, 0x00, 0x00000000, 0x00000000, 0x00);

	ret = wait_response(buf, 4, 1000);
	if (ret < 0)
		return -ERRLINE;

	dump_buffer("Load", buf, 4);

	ret = memcmp(response, buf, 4);
  
	return (ret ? -1 : 0);
}

static int
target_init(void)
{
	u8 buf[256];
	int ret;

	/* LED OFF */
	send_cmd(0x0202, 0x53fd0004, 0x20, 0x00000000, 0x00020000, 0x00);
	ret = wait_response(buf, 8, 1000);
	if (ret < 0)
		return -ERRLINE;

	dump_buffer("LED OFF Result", buf, 8);

	return 0;
}

static int
target_sync(void)
{
	u8 buf[256];
	int ret;
	int retry = 1024;
	u8 response[4] = {0x1e, 0x1e, 0x1e, 0x1e};

	message("Waiting for target - press Wakeup now.\n");
	while (1) {
		while (1) {
			if (retry-- < 0)
				return -ERRLINE;

			if (shoehorn_cancel_check()) {
				return -1;
			}
      
			write_serial_byte(0x00, 1);
			ret = wait_response(&buf[0], 1, 10);
			if (ret < 0)
				continue;
      
			ret += wait_response(&buf[1], 3, 10);
			if (ret == 4)
				break;
		}
		dump_buffer("Negotiation", buf, 4);
    	message("Initializing target... \n");

		send_cmd(0x0505, 0x00000000, 0x00, 0x00000000,
			 0x00000000, 0x00);
		ret = wait_response(buf, 4, 100);
		if (ret < 0)
			return -ERRLINE;
    
		dump_buffer("Sync Result", buf, 4);
		if (memcmp(response, buf, 4) == 0)
			break;
	}

	return 0;
}

static int
loader_wait_wakeup_armadillo5x0(void)
{
	int ret;

	serial_baud(115200);
	ret = target_sync();
	if (ret < 0) {
		if (ret != -1) {
			sprintf(msg, "Failed Target Synchronizing.(%d)\n", ret);
			message(msg);
			messagebox(msg);
		}
		serial_baud(9600);
		return -1;
	}

	ret = target_init();
	if (ret < 0) {
		sprintf(msg, "Failed Target Initializing.(%d)\n", ret);
		message(msg);
		messagebox(msg);
		serial_baud(9600);
		return -1;
	}

	return 0;
}

static int
loader_load_armadillo5x0(u8 *buf, u32 size)
{
	int ret;

	serial_baud(115200);
	ret = target_load_image(buf, size);
	if (ret < 0) {
		sprintf(msg, "Failed Loading Image.(%d)\n", ret);
		message(msg);
		messagebox(msg);
		serial_baud(9600);
		return -1;
	}
	return 0;
}

static int
loader_wait_ready_armadillo5x0(void)
{
	return 0;
}

void
loader_ops_setup_armadillo5x0(struct loader_operations *ops)
{
	ops->wait_wakeup	= loader_wait_wakeup_armadillo5x0;
	ops->load		= loader_load_armadillo5x0;
	ops->wait_ready		= loader_wait_ready_armadillo5x0;
}
