/* 
* shoehorn.c --	Host-side bootstrap loader for Armadillo-9 boards.
* 
* Copyright (C) 2001 Atmark Techno, Inc.
* Copyright (c) 2000 Blue Mug, Inc.  All Rights Reserved.
* Copyright (c) 1999 Ben Williamson <benw@pobox.com>
*
* This program loads a piece of code (see loader.c) into the target's
* internal SRAM.  The loader then does our bidding to initialise and
* detect hardware, load a kernel, ramdisk image, and boot.
*
* To add support for a new board type, add a new command-line option
* that sets the hardware variable, and arrange for it to call
* init_myboardname() instead of init_anvil().
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "shoehorn.h"

#if !defined(WIN32)
#include <unistd.h>
#include <getopt.h>
#include "serial.h"
#else
#include "win32/getopt.h"
#include "win32/serial_win.h"
#endif


#include "util.h"
#include "target.h"
#include "log.h"
#include "operation.h"

#define _POSIX_SOURCE	1

#define MAKE_VERSION(ver) "shoehorn v" #ver " built " \
	__TIME__", "__DATE__ \
" (CVS: $Id: shoehorn.c,v 1.10 2001/01/17 22:50:56 miket Exp $)";
static const char version[] 	= MAKE_VERSION(3.4-pre1);


#define DEFAULT_SRAM_SIZE	(0x800) /* 2kB of SRAM for loader.bin */
#define MAX_FRAGS	100
#define DEFAULT_DRAM_START	0xc0000000
static unsigned long dram_start;

#define MINORBITS	8
#define MKDEV(ma,mi)	(((ma) << MINORBITS) | (mi))
#define RAMDISK_MAJOR	1

#define PARAM_OFFSET	0x00020000	/* immediately after video RAM */
#define PARAM_U1_OFFSET	0x00020000
#define PARAM_U2_OFFSET	0x00020100
#define PARAM_CMDLINE_OFFSET	0x00020500
#define PARAM_END		0x00020900
#define PARAM_SIZE		(256 + 1024 + 1024)

#define DEFAULT_KERNEL_OFFSET	0x00218000	/* beginning of kernel image */
#define PAGE		0x1000

#define DEFAULT_INITRD_START	0xc0400000

#if !defined(WIN32)
#define ETH_STEP	1024
#endif

static int terminal = 0;
static int boot     = 0;

struct option options[] = {
	{ "boot",	0, &boot,	1 },
	{ "initrd",	1, 0,		'i' },
	{ "kernel",	1, 0,		'k' },
	{ "loader",	1, 0,		'l' },
	{ "initfile",   1, 0,           'I' },
	{ "postfile",   1, 0,           'P' },
	{ "target",	1, 0,		't' },
#if !defined(WIN32)
	{ "netif",	1, 0,		'n' },
#endif
	{ "port",	1, 0,		'p' },
	{ "terminal",	0, &terminal,	1 },
	{ "version",	0, 0,		'v' },
	{ 0,		0, 0,		0 }
};

/* this magic is documented in the GNU cpp info files; we use
it here to let a Makefile or build script specify the path
to loader.bin. */
#ifdef LOADERPATH
#define loaderpath(s) str(s)
#else
#define loaderpath(s)
#endif
#define str(s) #s

#if !defined(WIN32)
#define getopt_init() 
#endif

static char *initrd	= "initrd";
static char *kernel	= "/usr/lib/hermit/loader-armadillo9-boot.bin";
static char *loader	= loaderpath(LOADERPATH) "shoehorn-armadillo9.bin";
static char *initfile   = loaderpath(LOADERPATH) "shoehorn-armadillo9.init";
static char *postfile   = loaderpath(LOADERPATH) "shoehorn-armadillo9.post";
static char *target	= "armadillo9";
#if !defined(WIN32)
static char *netif	= "eth0";
#endif
static char *port	= "/dev/ttyS0";

char *progname		= "UNKNOWN";

char kargs[256];
unsigned char remotemac[6];

struct fragment {
	unsigned int	start;
	unsigned int	size;
} frag_list[MAX_FRAGS], *frag;

void display_version(void){
	char str[512];
	sprintf(str, "%s\n\n", version);
	message(str);
	messagebox(str);
}
/*****************************************************************************
* Print usage instructions and exit
*****************************************************************************/
void usage(void){
	char str[1024];
	int ptr = 0;
	ptr += sprintf(&str[ptr], "Usage: %s "
		"[options] [kernel command line]\n",progname);
	ptr += sprintf(&str[ptr], "Available options (defaults):\n");
	ptr += sprintf(&str[ptr], "        --boot\n");
	ptr += sprintf(&str[ptr], "        --initrd (%s)\n", initrd);
	ptr += sprintf(&str[ptr], "        --kernel (%s)\n", kernel);
	ptr += sprintf(&str[ptr], "        --loader (%s)\n", loader);
	ptr += sprintf(&str[ptr], "        --initfile (%s)\n", initfile);
	ptr += sprintf(&str[ptr], "        --postfile (%s)\n", postfile);
#if !defined(WIN32)
	ptr += sprintf(&str[ptr], "        --netif (%s)\n", netif);
#endif
	ptr += sprintf(&str[ptr], "        --port (%s)\n", port);
	ptr += sprintf(&str[ptr], "        --terminal\n");
	ptr += sprintf(&str[ptr], "        --version\n");
	message(str);
}

/*****************************************************************************
* Parse the command line options
*****************************************************************************/
static int parse_command_line(int argc, char **argv){
	int c;
	
	while (1) {
		c = getopt_long_only(argc, argv, "iklnp", options, NULL);
		if (c == -1) {
			break;
		}
		switch (c) {
		case 0:
			break;
		case 'i':
			initrd = optarg;
			break;
		case 'k':
			kernel = optarg;
			break;
		case 'l':
			loader = optarg;
			break;
		case 'I':
			initfile = optarg;
			break;
		case 'P':
			postfile = optarg;
			break;
		case 't':
			target = optarg;
			break;
#if !defined(WIN32)
		case 'n':
			netif = optarg;
			break;
#endif
		case 'p':
			port = optarg;
			break;
		case 'v':
			display_version();
			return -1;
		default:
			usage();
			return -1;
		}
	}
	return 0;
}

/*****************************************************************************
* concatenating what's left into kargs, but after root privs are dropped
*****************************************************************************/
static int build_kargs(int argc, char **argv){
	unsigned int remspace;
	unsigned int space;
	unsigned int prependspace = 0;	/* 0 on first arg, else 1 */
	
	memset(kargs, 0, sizeof kargs);
	while(optind < argc){
		remspace = sizeof(kargs) - strlen(kargs) - 1;
		space = strlen(argv[optind]) + prependspace;
		if(remspace >= space){
			/* will fit */
			if(prependspace){
				strcat(kargs, " ");
			}
			strcat(kargs, argv[optind++]);
		}else{
			message("Kernel options too long\n");
			messagebox("Kernel options too long\n");
			return -1;
		}
		prependspace = 1;
	}
	kargs[sizeof(kargs) - 1] = '\0';
	return 0;
}

/*****************************************************************************
*****************************************************************************/
int target_write(unsigned addr, const char *buf,
				 unsigned size, unsigned progress){
	/* XXX this is kind of nasty */
	target_write_block(addr, buf, size, progress);
	return 0;
}

/*****************************************************************************
* Write an object to the target, spanning multiple DRAM fragments
* Returns final address
*****************************************************************************/
unsigned int target_write_fragmented(unsigned int addr, char *buf, 
									 unsigned int size){
	struct fragment *frag;
	unsigned int frag_end = 0;
	int progress = 0;
	
	for(frag = &frag_list[1]; frag->size != 0; frag++){
		frag_end = frag->start + frag->size;
		if((frag->start <= addr) && (addr < frag_end)){
			break;
		}
	}
	while(size > 0){
		int step = min(size, frag_end - addr);
		if(step <= 0){
			message("Insufficient DRAM space\n");
			messagebox("Insufficient DRAM space\n");
			return -1;
		}
		target_write(addr, buf, step, progress);
		addr += step;
		buf += step;
		progress += step;
		size -= step;
		if(size > 0){
			frag++;
			addr = frag->start;
			frag_end = frag->start + frag->size;
		}
	}
	if(addr == frag_end){
		addr = frag[1].start;
	}
	return addr;
}

/*****************************************************************************
*****************************************************************************/
int ping(void){
	char c;
	char str[256];
	
	message("Pinging loader\n");
	
	put_char('a');
	c = get_char();
	if (c != '+') {
		sprintf(str, "Expected acknowledge '+' "
			"but got '%c' (0x%02x)\n",
			(c ? c : ' '), c);
		message(str);
		messagebox(str);
		return -1;
	}
	return 0;
}

/*****************************************************************************
* This definition of param_struct is copied from include/asm-arm/setup.h
* in Linux 2.4.0-test6-rmk5.  Changes made after that kernel should
* hopefully remain backwards compatible.
*
* Note that these will have to be changed to e.g. 'u32' to work on a 64-
* bit development workstation.
*****************************************************************************/
#define COMMAND_LINE_SIZE 1024

struct param_struct {
    union {
		struct {
			unsigned long page_size;		/*  0 */
			unsigned long nr_pages;		/*  4 */
			unsigned long ramdisk_size;		/*  8 */
			unsigned long flags;		/* 12 */
#define FLAG_READONLY	1
#define FLAG_RDLOAD	4
#define FLAG_RDPROMPT	8
			unsigned long rootdev;		/* 16 */
			unsigned long video_num_cols;	/* 20 */
			unsigned long video_num_rows;	/* 24 */
			unsigned long video_x;		/* 28 */
			unsigned long video_y;		/* 32 */
			unsigned long memc_control_reg;	/* 36 */
			unsigned char sounddefault;		/* 40 */
			unsigned char adfsdrives;		/* 41 */
			unsigned char bytes_per_char_h;	/* 42 */
			unsigned char bytes_per_char_v;	/* 43 */
			unsigned long pages_in_bank[4];	/* 44 */
			unsigned long pages_in_vram;	/* 60 */
			unsigned long initrd_start;		/* 64 */
			unsigned long initrd_size;		/* 68 */
			unsigned long rd_start;		/* 72 */
			unsigned long system_rev;		/* 76 */
			unsigned long system_serial_low;	/* 80 */
			unsigned long system_serial_high;	/* 84 */
			unsigned long mem_fclk_21285;       /* 88 */ 
		} s;
		char unused[256];
    } u1;
    union {
		char paths[8][128];
		struct {
			unsigned long magic;
			char n[1024 - sizeof(unsigned long)];
		} s;
    } u2;
    char commandline[COMMAND_LINE_SIZE];
};

/*****************************************************************************
*****************************************************************************/
int target_write_params(unsigned long initrd_start,
						unsigned long initrd_size){
	struct param_struct ps;
	char str[256];
	
	/* sanity checking to keep updates from breaking anything */
	if(PARAM_SIZE != PARAM_END - PARAM_OFFSET){
		return -1;
	}
	if(sizeof (struct param_struct) > PARAM_SIZE){
		return -1;
	}
	
	/* XXX these need to be generalized for different HW setups */
	/* XXX they are also endian specific */
	memset(&ps, 0, sizeof ps);
	/* printf("- page_size: %d\n", PAGE); */
	ps.u1.s.page_size = PAGE;
	
	sprintf(str, "- nr_pages (all banks): %d\n", 4096);
	message(str);
	
	ps.u1.s.nr_pages = 4096;
	/* XXX */
	ps.u1.s.flags = FLAG_READONLY | FLAG_RDLOAD | FLAG_RDPROMPT;
	
	message("- rootdev: (RAMDISK_MAJOR, 0)\n");
	
	ps.u1.s.rootdev = MKDEV(RAMDISK_MAJOR, 0);
	
	sprintf(str, "- pages_in_bank[0]: %d\n", 2048);
	message(str);
	
	ps.u1.s.pages_in_bank[0] = 2048;
	sprintf(str, "- pages_in_bank[1]: %d\n", 2048);
	message(str);
	
	ps.u1.s.pages_in_bank[1] = 2048;
	
	sprintf(str, "- initrd_start: 0x%lx\n", initrd_start);
	message(str);
	
	ps.u1.s.initrd_start = initrd_start;
	
	sprintf(str, "- initrd_size: 0x%lx\n", initrd_size);
	message(str);
	
	ps.u1.s.initrd_size = initrd_size;
	
	sprintf(str, "- ramdisk_size: 0x%lx\n", initrd_size);
	message(str);
	
	ps.u1.s.ramdisk_size = initrd_size;
	
	/* XXX these writes overlap */
	print_size(dram_start + PARAM_OFFSET, PARAM_SIZE);
	target_write_fragmented(dram_start + PARAM_OFFSET,
		(char*) &ps, PARAM_SIZE);
	target_write_fragmented(dram_start + PARAM_CMDLINE_OFFSET,
		kargs, sizeof(kargs));
	return 0;
}

/*****************************************************************************
*****************************************************************************/
int detect_dram(void){
	unsigned int start, size, total_size;
	char str[256];
	
	put_char('d');
	
	sprintf(str, "- %d bits wide\n", get_char());
	message(str);
	
	frag_list[0].start = 0;	/* item 0 is a dummy, list starts at 1 */
	frag_list[0].size = 0;
	total_size = 0;
	frag = frag_list;
	size = get_word();

	start = get_word();
	do{
		if(start == frag->start + frag->size){
			frag->size += size;
		}else{
			frag++;
			if(frag >= &frag_list[MAX_FRAGS-1]){
				message("Too many DRAM fragments\n");
				messagebox("Too many DRAM fragments\n");
				return -1;
			}
			frag->start = start;
			frag->size = size;
		}
		total_size += size;

		start = get_word();
		if(start == 0){
			break;
		}
	}while(1);

	frag++;
	frag->start = frag->size = 0;
	for(frag = &frag_list[1]; frag->size != 0; frag++){
		print_size(frag->start, frag->size);
	}
	
	sprintf(str, "Total DRAM: %dkB\n", total_size / 1024);
	message(str);
	return 0;
}

/*****************************************************************************
*****************************************************************************/
void (*shoehorn_start_callback)(void) = (void (*)(void))NULL;
#define shoehorn_start() \
	if(shoehorn_start_callback != NULL) \
shoehorn_start_callback()

/*****************************************************************************
*****************************************************************************/
static int cancel_flag = 0;
void shoehorn_cancel(void){
	cancel_flag = 1;
}

int 
shoehorn_cancel_check(void)
{
	return cancel_flag;
}

/*****************************************************************************
*****************************************************************************/
void set_program_name(char *path){
	progname = strrchr(path, '/');
	if(progname == NULL){
		progname = path;
	}else{
		progname++;
	}
	return;
}

/*****************************************************************************
*****************************************************************************/
int
setup_configuration(char *conf_file, void *func, u32 *arg, u32 defval)
{
	u32 (*callback)(FILE *) = func;
	FILE *fp;

	fp = fopen(conf_file, "r");
	if (fp == NULL)
		return -1;

	*arg = callback(fp);
	fseek(fp, 0L, SEEK_SET);
	fclose(fp);
  
	if (*arg == -1)
		*arg = defval;

	return 0;
}
/*****************************************************************************
* Send loader.bin to the target
* Initialise some target registers, for Anvil hardware
* Detect what DRAM is present, building a list of fragments
* Load the kernel image and initrd into available fragments
* Write a kernel parameter block with command-line arguments
* Jump to the kernel entry point
*****************************************************************************/
#if !defined(WIN32)
#define MAIN_FANCTION main
#else
#define MAIN_FANCTION shoehorn_main
#endif

int MAIN_FANCTION(int argc, char **argv){
	unsigned char *kernel_buf, *loader_buf;
	
	int kernel_size, initrd_size, loader_size;
	
	unsigned long kernel_end, initrd_start, kernel_offset;
	u32 sram_size;
	int size;
#if !defined(WIN32)
	unsigned char *initrd_buf;
	uid_t ruid, euid, suid;
	int getresuid(uid_t *, uid_t *, uid_t *);	/* linux only????? */
#endif
	
	FILE *fp;
	int ret;
	char str[256];
	
	cancel_flag = 0;
	set_program_name(argv[0]);
	
	getopt_init();
	ret = parse_command_line(argc, argv);
	if(ret == -1){
		return -1;
	}
	
	/* loader_operations setup */
	loader_ops_init(target);

	/* make output to stdout immediately visible */
	setbuf(stdout, NULL);
	
#if !defined(WIN32)
	/* drop root privs if setuid root */
	if ( (geteuid() == 0) && (getuid() != 0) ) {
		message("Dropping root privileges\n");
		if (setuid(getuid()) != 0) {
			sprintf(str, "%s: error in droping root-privs\n",
				progname);
			message(str);
			exit(EXIT_FAILURE);
		}
		/* FIXME: make it portable to systems w/o saved uid */
		/* check if root privs are really dropped */
		if (getresuid(&ruid, &euid, &suid) != 0) {
			sprintf(str, "%s: could not check if root-privs "
				"were dropped\n", progname);
			message(str);
			exit(EXIT_FAILURE);
		}
		if ( (ruid == 0) || (euid == 0) || (suid == 0) ) {
			sprintf(str, "%s: not all root-privs were dropped\n"
				"\truid: %d, euid: %d, suid: %d\n",
				progname, ruid, euid, suid);
			message(str);
			exit(EXIT_FAILURE);
		}
	}
#endif
	
	/* fill in kargs *after* dropping privileges */
	ret = build_kargs(argc, argv);
	if(ret == -1){
		return -1;
	}

	/* Load Configuration */
	ret = setup_configuration(initfile, get_sram_size,
				  &sram_size, DEFAULT_SRAM_SIZE);
	if (ret < 0) {
		print_perror(initfile);
		return -1;
	}
	setup_configuration(initfile, get_dram_start,
			    &dram_start, DEFAULT_DRAM_START);
	setup_configuration(initfile, get_initrd_start,
			    &initrd_start, DEFAULT_INITRD_START);
	setup_configuration(initfile, get_kernel_offset,
			    &kernel_offset, DEFAULT_KERNEL_OFFSET);
	
	/* slurp files into buffers */
	loader_size = sram_size;
	/* must allocate at least sram_size bytes */
	ret = read_file(loader, (void **)&loader_buf, &loader_size);
	if(ret == -1) return -1;
	
	kernel_size = 0;
	ret = read_file(kernel, (void **)&kernel_buf, &kernel_size);
	if(ret == -1) return -1;
	
	initrd_size = 0;
#if !defined(WIN32)
	ret = read_file(initrd, (void **)&initrd_buf, &initrd_size);
	if(ret == -1) return -1;
#endif
	
	/* make sure loader isn't too big */
	if(loader_size > sram_size){
		sprintf(str, "%s: loader too large (limit %d bytes)\n",
			progname, sram_size);
		message(str);
		messagebox(str);
		return -1;
	}
	
	/* open serial port and start talking to hardware */
	ret = serial_open(port);
	if(ret == -1) return -1;

	ret = loader_ops.wait_wakeup();
	if (ret < 0) {
		serial_close();
		return -1;
	}
	shoehorn_start();

	message("Writing SRAM loader...\n");
	loader_ops.load(loader_buf, sram_size);
	ret = loader_ops.wait_ready();
	if (ret < 0) {
		sprintf(str, "Writeing SRAM loader uncompreted.\n");
		message(str);
		messagebox(str);
		serial_close();
		return -1;
	}

	ret = ping();
	if(ret == -1){
		serial_close();
		return -1;
	}
	
	message("Initialising hardware:\n");

	fp = fopen(initfile, "r");
	execute_script(fp,0);
	fclose(fp);
	
	ret = ping();
	if(ret == -1){
		serial_close();
		return -1;
	}
	
	message("Detecting DRAM\n");
	detect_dram();
	
	sprintf(str, "Loading %s:\n", kernel);
	message(str);
	
	print_size(dram_start + kernel_offset, kernel_size);
	kernel_end = target_write_fragmented(dram_start + kernel_offset,
		kernel_buf, kernel_size);
	free(kernel_buf);
	
	/* Find a start address for initrd.  We put it
	   at the highest possible page-aligned address. */
	size = (initrd_size + PAGE - 1) & ~(PAGE - 1);
	for (frag = &frag_list[1]; frag->size != 0; frag++);
	frag--;
	while ((size > 0) && (frag > frag_list)) {
		int step = min(size, (int)frag->size);
		initrd_start = frag->start + frag->size - step;
		size -= step;
		frag--;
	}
	/* XXX miket: trying this out */
	sprintf(str, "initrd_start is %lx\n", initrd_start);
	message(str);
	
	sprintf(str, "Moving initrd_start to %lx\n", initrd_start);
	message(str);
	
	/* end XXX area */
	if (initrd_start < kernel_end) {
		message("Not enough space for initrd\n");
		messagebox("Not enough space for initrd\n");
		serial_close();
		return -1;
	}
	
	sprintf(str, "Loading %s:\n", initrd);
	message(str);
	
#if !defined(WIN32)
	print_size(initrd_start, initrd_size);
	target_write_fragmented(initrd_start, initrd_buf, initrd_size);
	free(initrd_buf);
#endif
	
	message("Writing parameter area\n");
	
	target_write_params(initrd_start, initrd_size);
	
	fp = fopen(postfile, "r");
	if(fp == NULL){
		print_perror(postfile);
		serial_close();
		return -1;
	}
	execute_script(fp, 0);
	fclose(fp);
	
	ret = ping();
	if(ret == -1){
		serial_close();
		return -1;
	}
	
	if (boot) {
		sprintf(str, "Starting kernel at 0x%08lx\n",
			dram_start + kernel_offset);
		message(str);

		put_char('c');
		put_word(dram_start + kernel_offset);
		put_word(0);	/* sanity check */
		put_word(0);
		put_word(0);
		put_word(0);
	}
	
#if !defined(WIN32)
	/* I've found this bit useful for debugging: If the kernel doesn't
	   seem to boot, run "shoehorn --terminal" and plug head-armv.S
	   and init/main.c full with things like "IO_UARTDR1 = '!';" or
	   similar.  Then you'll find out how far the setup code gets. */
	if (terminal)
		serial_terminal();
#endif
	
	serial_close();
	return 0;
}

